from p10s.values import value
from pathlib import Path


class BaseContext():
    def __init__(self, input=None, output=None):
        if input is None:
            self.input = value('p10s', {}).get('file', None)
        else:
            self.input = Path(input)

        if output is not None:
            self.output = Path(output)
        elif self.input is not None:
            self.output = self.input.with_suffix(self.output_file_extension)
        else:
            self.output = None

    def render(self):
        raise NotImplementedError()  # pragma: no cover

    def _output_stream(self):
        self.output.parent.mkdir(parents=True, exist_ok=True)
        return self.output.open("w")

    def __repr__(self):
        return "<" + self.__module__ + "." + self.__class__.__name__ + ">"



def merge_dicts(*args):
    """Creates a new dict by merging together the values in
``args``. Values to the "right" over ride values in the "left"."""
    if len(args) == 0:
        return {}
    elif len(args) == 1:
        return args[0]
    else:
        args = list(args)
        first = args.pop(0)
        return _merge_dicts(first, merge_dicts(*args))


def _merge_dicts(a, b):
    """Copy values from B into A. destructively modifies A."""
    def rec(a, b):
        for k in b.keys():
            new = b[k]
            if isinstance(new, dict):
                existing = a.get(k, a)
                if existing is a:
                    a[k] = new
                else:
                    if isinstance(existing, dict):
                        rec(existing, new)
                    else:
                        a[k] = new
            else:
                a[k] = new

    rec(a, b)

    return a



from pathlib import Path
from collections.abc import MutableMapping
from copy import deepcopy
from contextlib import contextmanager
import os

from p10s.utils import merge_dicts
from p10s.loads import load_file


class Values(MutableMapping):
    """Class for storing p10s value mappings.

Can be used as a dict. Has classmethods, such as
:py:meth:`p10s.values.Values.from_files` for loading values from
various sources.

The assumption is that Values won't be created directly, though that's
certainly possible, but that they're be created from external
source. If, for example, we want to have our values stored in a yaml
on disk, and allow over riding form the env, we would this:

.. code-block:: python

    VALUES = Values.from_files(".") + Values.from_environ(".")


    """
    def __init__(self, values=None):
        if values is not None:
            self.values = values
        else:
            self.values = {}

    @classmethod
    def from_files(cls, basedir):
        """Builds a Values object from a hierarchically organzied collections
of yaml files.

All the files named ``values.yaml`` in basedir and up the file system
(up until the file system's root) will be collected and parsed and
merged. The merging is top down, so values specifed in files closer to
basedir will over ride values specified in a higher up values file.
        """
        basedir = Path(basedir)
        if not basedir.exists():
            raise FileNotFoundError("basedir %s does not exist" % basedir)
        here = basedir

        values_files = []

        if basedir.is_file():
            basedir = basedir.parent()

        while True:
            if (here / 'values.yaml').exists():
                values_files.insert(0, here / 'values.yaml')
            if here == here.parent:
                break
            else:
                here = here.parent

        values = {}
        for file in values_files:
            merge_dicts(values, load_file(file))

        return cls(values)

    @classmethod
    def from_environ(cls):
        """Builds a Values object from the current OS environment."""
        # NOTE we may be able to just pass in os.enviro directly,
        # without creating a dict, but i'm not sure what other magic
        # is on that and this feels safer. 20181220:mb
        return cls(dict(os.environ))

    def __add__(self, other):
        """Returns a new values containg the merge of ``other`` into this
        object (key/value pairs in ``other`` replace equally named
        pairs in ``self``)"""
        new = self.copy()
        return new.__iadd__(other)

    def __iadd__(self, other):
        """Modifies ``self`` by merging in the values of ``other``"""
        self.values = merge_dicts(self.values, other)
        return self

    def copy(self):
        return Values(values=deepcopy(self.values))

    def __getitem__(self, key):
        return self.values[key]

    def __setitem__(self, key, value):
        self.values[key] = value

    def __delitem__(self, key):
        del self.values[key]

    def __iter__(self):
        return iter(self.values)

    def __len__(self):
        return len(self.values)

    def __keytransform__(self, key):
        return key

    def set_value(self, key, value):
        self.values[key] = value

    def get_value(self, key, default=None):
        return self.values.get(key, default)


global VALUES
VALUES = Values()


def use_values(values):
    global VALUES
    VALUES = values


@contextmanager
def values(*args, **kwargs):
    """context manager, runs the body with the given key=value pairs
bound.

``*args`` is a list of dicts, whose final values will be computed as
by :func:``merge_dicts <p10s.utils.merge_dicts>`. ``*kwargs`` are
tweated as a single dict of key=value pairs, and take precedence over
anything in ``*args``.

    """
    global VALUES
    old = VALUES
    new = old.copy()

    for dict in args:
        new += Values(values=dict)
    new += Values(values=kwargs)

    VALUES = new
    yield
    VALUES = old


def value(key, default=None):
    """Returns the value of the key named ``key``."""
    global VALUES
    return VALUES.get_value(key, default)


def set_value(key, value):
    """Sets the value with the key ``key`` to ``value``."""
    global VALUES
    VALUES.set_value(key, value)
    return value



"""It is often easier to write some terraform or yaml code directly in
the native syntax instead of having to write out the full python
object using python's literal dict/list syntax.

Note: you probably shouldn't call these functions directly, while they
document the exact syntax and haviour of the parsers, you probably
want to be using one of the specific object generators instead:
:func:`from_hcl <p10s.terraform.from_hcl>`, :func:`many_from_hcl
<p10s.terraform.many_from_hcl>`, :func:`from_yaml
<p10s.kubernetes.from_yaml>` and :func:`many_from_yaml
<p10s.kubernetes.many_from_yaml>`.

.. autofunction:: p10s.loads.yaml
.. autofunction:: p10s.loads.yaml_all
.. autofunction:: p10s.loads.hcl
.. autofunction:: p10s.loads.json

"""
from pathlib import Path
from ruamel.yaml import YAML
import hcl as pyhcl
import json as json_lib
import io


ruamel = YAML(typ='safe', pure=True)
ruamel.default_flow_style = False


def _data(object):
    if isinstance(object, (io.BufferedIOBase, io.TextIOBase, io.RawIOBase, io.IOBase)):
        return object.read()
    elif isinstance(object, Path):
        return object.open().read()
    else:
        return object


def yaml(input):
    """Parses ``input`` as a single yaml document and returns the
corresponding python data structure.

:param input: the source of the yaml
:type input: str, Path or IOBase
    """
    try:
        return ruamel.load(_data(input))
    except Exception as e:
        print("Error parsing %s" % input)
        raise(e)


def yaml_all(input):
    """Parses ``input`` as multiple yaml documents and returns a list of python objects (dicts, lists, etc.)

:param input: the source of the yaml
:type input: str, Path or IOBase
"""
    try:
        return ruamel.load_all(_data(input))
    except Exception as e:
        print("Error parsing %s" % input)
        raise(e)


def hcl(input):
    """Parses ``input`` as a hcl code and returns the corresponding python dict.

As with terraform's json syntax blocks are converted to nested dicts:

.. code-block:: text

    terraform {
      a = "b"
    }

parses to

.. code-block:: python

    {'terraform':
        {'a': 'b'}
    }

and this:

.. code-block:: terraform

    resource "type" "name" {
      a = "b"
    }

parses to

.. code-block:: python

    {'resource':
        {'type':
            {'name':
                {'a': 'b'}
            }
        }
    }

Multiple blocks are merged into a single dict as per the tf.json format:

.. code-block:: terraform

    resource "type" "first" {
      a = "b"
    }
    resource "type" "second" {
      a = "b"
    }

parses to

.. code-block:: python

    {'resource':
        {'type':
            {'first':  {'a': 'b'}},
            {'second': {'a': 'b'}},
        }
    }

:param input: the source of the hcl
:type input: str, Path or IOBase
"""
    return pyhcl.loads(_data(input))


def json(input):
    """Parses ``input`` as a single json object.

:param input: the source of the json
:type input: str, Path or IOBase
"""
    return json_lib.loads(_data(input))


def load_file(filename):
    filename = Path(filename)
    suffix = filename.suffix
    if suffix == '.yaml':
        return yaml(filename)
    elif suffix == '.hcl':
        return hcl(filename)
    elif suffix == '.json':
        return json(filename)
    else:
        raise Exception('Unknown extension {}' % filename)



"""The ``p10s`` module doesn't contain any actual code, is exists to
provide a convenient way to import code into a p10s script.

Example:

.. code-block:: python

    from p10s import tf, k8s

The following things are importable from this module:

``tf``
    just the ``p10s.terraform`` module
``k8s``
    the ``p10s.kubernetes`` module
``values``
    the ``p10s.values`` module
``yaml``, ``json``, ``hcl``
    parsers for syntax of the given type

"""

from p10s.loads import yaml, json, hcl  # noqa: F401

import p10s.kubernetes as k8s  # noqa: F401
import p10s.terraform as tf  # noqa: F401
import p10s.values  # noqa: F401
from p10s.__version__ import __version__  # noqa: F401

from p10s.generator import register_context



import sys
import os
from pathlib import Path
import runpy
import copy
from contextlib import contextmanager
from pprint import pformat

from p10s.base import BaseContext
from p10s.values import values


@contextmanager
def _global_state(dir, extra_sys_paths):
    here = os.getcwd()
    sys_path = copy.copy(sys.path)
    try:
        os.chdir(str(dir))
        sys.path = [str(path) for path in extra_sys_paths] + sys.path
        yield
    finally:
        os.chdir(here)
        sys.path = sys_path


CONTEXTS = []


def register_context(context):
    CONTEXTS.append(context)


def _stderr(*args):
    print(*args, file=sys.stderr)


class P10SScript():
    def __init__(self, filename):
        self.filename = filename
        self.base_dir = filename.parent
        self.contexts = []
        self.pyterranetes_dir = self._find_pyterranetes_dir(filename.parent)

    def render(self, verbose=False):
        for c in self.contexts:
            with _global_state(dir=self.base_dir,
                               extra_sys_paths=[self.pyterranetes_dir]):
                if verbose:
                    _stderr("  Rendering", pformat(c), "to", c.output, "in", self.base_dir)
                c.render()
        return self

    def compile(self, verbose=False):
        if verbose:
            _stderr("Compiling", self.filename)
        with values({'p10s': {'file': self.filename}}):
            with _global_state(dir=self.base_dir,
                               extra_sys_paths=[self.pyterranetes_dir]):
                CONTEXTS.clear()
                globals = runpy.run_path(str(self.filename))
                for value in globals.values():
                    if isinstance(value, BaseContext):
                        self.contexts.append(value)
                self.contexts.extend(CONTEXTS)
        return self

    def _find_pyterranetes_dir(self, root):
        here = root
        count = 0
        while True:
            maybe = here / 'pyterranetes'
            if maybe.exists() and maybe.is_dir():
                return maybe
                break
            if str(here) == here.root:
                break
            here = here.parent

            if count > 1000:
                raise Exception("Directory tree too deep. aborting")
            count += 1


class Generator():

    def _p10s_scripts(self, root):
        if not root.exists():
            raise FileNotFoundError("%s does not exist." % str(root))
        if root.is_file():
            yield root
        else:
            for dirname, sub_dirs, files in os.walk(str(root)):
                for file in files:
                    path = Path(file)
                    if path.suffix == '.p10s':
                        yield Path(os.path.join(dirname, file))

    def generate(self, root, verbose=False):
        root = Path(root).resolve()
        scripts = [P10SScript(filename=filename) for filename in self._p10s_scripts(root)]
        for script in scripts:
            try:
                script.compile(verbose=verbose).render(verbose=verbose)
            except Exception as e:
                if verbose:
                    _stderr("Error while generating %s", script.filename)
                raise e
